// ===================================================================
//
// CBP_Sub.c
// PID controller to move sub to a desired height above floor
// CBP 03-05-21
//
// ==================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <stdio.h>
#include <webots/light_sensor.h>
#include <webots/gps.h>

#include "CBP_Helper_3.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 1.0
#define PI 3.14159265358979

WbDeviceTag left_motor, right_motor, back_left_motor, back_right_motor;
WbDeviceTag ls0, ls7;
WbDeviceTag gps;

// PID Stuff for thrust -------------------------------------------------
// Declare variables --------------------------------
double e1_thrust = 0.0;
double e2_thrust = 0.0;
double output_thrust = 0.0;
double error_thrust = 0.0;

// reset controller ---------------------------------
void reset_PID_control_thrust(){
 e1_thrust = 0.0;
 e2_thrust = 0.0;
 output_thrust = 0.0;
 error_thrust = 0.0;
}

// The thrust controller ----------------------------
double do_PID_control_thrust(double input, double target, double Kp, double Ki, double Kd, double dT){
  
  double Ts,du;
  Ts = dT;
  e2_thrust = e1_thrust;
  e1_thrust = error_thrust;
  
  error_thrust = target - input;
  du = Kp * (error_thrust - e1_thrust) + (Ki * Ts * error_thrust) + (Kd / Ts) * (error_thrust - 2*e1_thrust + e2_thrust);
  output_thrust += du;
  
  return output_thrust;
}
 
void deviceInit() {
  left_motor = wb_robot_get_device("motor1");
  right_motor = wb_robot_get_device("motor2");
  back_right_motor = wb_robot_get_device("motor3");
  back_left_motor = wb_robot_get_device("motor4");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_position(back_left_motor, INFINITY);
  wb_motor_set_position(back_right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);
  wb_motor_set_velocity(back_left_motor, 0.0);
  wb_motor_set_velocity(back_right_motor, 0.0);

  gps = wb_robot_get_device("gps1");
  wb_gps_enable(gps, TIME_STEP); 
       
}  



int main(int argc, char **argv) {

  double omegaL, omegaR;
  double driveL, driveR; 
  double Kp, Ki, Kd, dT;
  
 // double target_x;
 // double target_y;
  double target_z;
   
  double thrust = 0.0;
  
  wb_robot_init();
 
  deviceInit();

  reset_PID_control_thrust(); 
  
  target_z = 0.5;
  
  // Discover the coefficients based on experience
  Kp = 5.0;
  Kd = 1.1;
  Ki = 0.5;
  
  dT = (float)TIME_STEP/1000.0;
  
  while (wb_robot_step(TIME_STEP) != -1) {
          
    const double *gps_vals = wb_gps_get_values(gps);
    printf("gps[0] = %f, gps[1] = %f, gps[2] = %f\n",gps_vals[0],gps_vals[1],gps_vals[2]);
    
    thrust = 0.0 + 1.0*do_PID_control_thrust(gps_vals[1], target_z, Kp, Ki, Kd, dT);
      
    driveL = thrust;
    driveR = thrust;
    
    omegaL = driveR*MAX_SPEED;
    omegaR = driveL*MAX_SPEED;  
       
    printf("OmegaL: %f, omegaR: %f\n",omegaL,omegaR);
   
    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
    wb_motor_set_velocity(back_right_motor, omegaR);
    wb_motor_set_velocity(back_left_motor, omegaL);

  };

  wb_robot_cleanup();

  return 0;
}
