// =============================================================================
//
// FBO_Quadrotor_Devel_1.c
//
// =============================================================================

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <webots/robot.h>
 
#include <webots/camera.h>
#include <webots/compass.h>
#include <webots/gps.h>
#include <webots/gyro.h>
#include <webots/inertial_unit.h>
#include <webots/keyboard.h>
#include <webots/led.h>
#include <webots/motor.h>

//#include "CBP_Helper_Q.h"

#define SIGN(x) ((x) > 0) - ((x) < 0)
#define CLAMP(value, low, high) ((value) < (low) ? (low) : ((value) > (high) ? (high) : (value)))

// charactersitic of the quadrotor
//#define Ixx 0.10
#define Ixx 1
#define Iyy 1
#define Izz 1
#define h 0.305
#define l 0.292

#define A (h/(2*Ixx))
#define B (l/(2*Iyy))

#define I_THRESH 200

double e1_pitch = 0.0;
double e2_pitch = 0.0;
double output_pitch = 0.0;

double e1_roll = 0.0;
double e2_roll = 0.0;
double output_roll = 0.0;

double e1_yaw = 0.0;
double e2_yaw = 0.0;
double output_yaw = 0.0;

void reset_PID_control_pitch(){
 e1_pitch = 0.0;
 e2_pitch = 0.0;
 output_pitch = 0.0;
}

void reset_PID_control_yaw(){
 e1_yaw = 0.0;
 e2_yaw = 0.0;
 output_yaw = 0.0;
}

void reset_PID_control_roll(){
 e1_roll = 0.0;
 e2_roll = 0.0;
 output_roll = 0.0;
}

double errorP = 0;
double errorY = 0;
double errorR = 0;

double do_PID_control_pitch(double input, double target, double Kp, double Ki, double Kd, double dT){
  
  double Ts,du;
  Ts = dT;
  
  e2_pitch = e1_pitch;
  e1_pitch = errorP;
  errorP = target - input;
  du = Kp*(errorP - e1_pitch) + Ki*Ts*errorP + (Kd/Ts)*(errorP - 2*e1_pitch + e2_pitch);
  output_pitch += du;
  
  return output_pitch;
}


double do_PID_control_roll(double input, double target, double Kp, double Ki, double Kd, double dT){
  
  double Ts,du;
  Ts = dT;
  
  e2_roll = e1_roll;
  e1_roll = errorR;
  errorR = target - input;
  du = Kp*(errorR - e1_roll) + Ki*Ts*errorR + (Kd/Ts)*(errorR - 2*e1_roll + e2_roll);
  output_roll += du;
  
  return output_roll;
}


double do_PID_control_yaw(double input, double target, double Kp, double Ki, double Kd, double dT){
  
  double Ts,du;
  Ts = dT;
  
  e2_yaw = e1_yaw;
  e1_yaw = errorY;
  errorY = target - input;
  du = Kp*(errorY - e1_yaw) + Ki*Ts*errorY + (Kd/Ts)*(errorY - 2*e1_yaw + e2_yaw);
  output_yaw += du;
  
  return output_yaw;
}

void thrustToRotorSpeed(double thrust, double torquePitch, double torqueRoll, double torqueYaw,
                        double *rotor0, double *rotor1, double *rotor2, double *rotor3) {

  double r0 = 0;
  double r1 = 0;
  double r2 = 0;
  double r3 = 0;
  double k = 1;
  
  r0 -= (1/A)*torquePitch;
  r0 += (1/B)*torqueRoll;
  r0 -= torqueYaw;
  r0 += thrust;
  
  r1 += (1/A)*torquePitch;
  r1 += (1/B)*torqueRoll;
  r1 += torqueYaw;
  r1 += thrust;
    
  r2 += (1/A)*torquePitch;
  r2 -= (1/B)*torqueRoll;
  r2 -= torqueYaw;
  r2 += thrust;
    
  r3 -= (1/A)*torquePitch;
  r3 -= (1/B)*torqueRoll;
  r3 += torqueYaw;
  r3 += thrust;  
  
  printf("r0 ... r3 = (%f, %f, %f, %f)\n",r0,r1,r2,r3);
  
  *rotor0 = k*sqrt(fabs(r0))*(r0/fabs(r0));
  *rotor1 = k*sqrt(fabs(r1))*(r1/fabs(r1));
  *rotor2 = k*sqrt(fabs(r2))*(r2/fabs(r2));
  *rotor3 = k*sqrt(fabs(r3))*(r3/fabs(r3));
        
}

int main(int argc, char **argv) {

  double thrust;
  double FL,FR,RL,RR;
  double Kp, Ki, Kd, dT;
   
  FILE* logFile;
  char fname[10];
  const char* name;
  
  Kp = atof(argv[1]);
  Kd = atof(argv[2]);
  Ki = atof(argv[3]);  
  
  printf("Kp = %f, Kd = %f, Ki = %f\n",Kp,Kd,Ki);
  
  wb_robot_init();
  int timestep = (int)wb_robot_get_basic_time_step();
  printf("Timestep = %d\n",timestep);
  
  // Get and enable devices.
  WbDeviceTag camera = wb_robot_get_device("camera");
  wb_camera_enable(camera, timestep);
  WbDeviceTag front_left_led = wb_robot_get_device("front left led");
  WbDeviceTag front_right_led = wb_robot_get_device("front right led");
  WbDeviceTag imu = wb_robot_get_device("inertial unit");
  wb_inertial_unit_enable(imu, timestep);
  WbDeviceTag gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, timestep);
  WbDeviceTag compass = wb_robot_get_device("compass1");
  wb_compass_enable(compass, timestep);
  WbDeviceTag gyro = wb_robot_get_device("gyro");
  wb_gyro_enable(gyro, timestep);
  wb_keyboard_enable(timestep);
  WbDeviceTag camera_roll_motor = wb_robot_get_device("camera roll");
  WbDeviceTag camera_pitch_motor = wb_robot_get_device("camera pitch");
  // WbDeviceTag camera_yaw_motor = wb_robot_get_device("camera yaw");  // Not used in this example.

  // Get propeller motors and set them to velocity mode.
  WbDeviceTag front_left_motor = wb_robot_get_device("front left propeller");
  WbDeviceTag front_right_motor = wb_robot_get_device("front right propeller");
  WbDeviceTag rear_left_motor = wb_robot_get_device("rear left propeller");
  WbDeviceTag rear_right_motor = wb_robot_get_device("rear right propeller");
  WbDeviceTag motors[4] = {front_left_motor, front_right_motor, rear_left_motor, rear_right_motor};
 
  int m;
  for (m = 0; m < 4; ++m) {
    wb_motor_set_position(motors[m], INFINITY);
    wb_motor_set_velocity(motors[m], 1.0);
  }

  wb_keyboard_enable(timestep);

  // Wait one second.
  while (wb_robot_step(timestep) != -1) {
    if (wb_robot_get_time() > 1.0)
      break;
  }

  name = wb_robot_get_name();
  sprintf(fname,"Robot_%s.m",name);
  logFile = fopen(fname,"w");
  //writeOctaveHead(logFile);
  
  // (Task-6) generate control signal - targets
  double target_pitch = 0.0;
  double target_roll = 0.0;
  double target_yaw = 0.0;   

  // (Task-4) testing the thrust to speed function
  //tauPitch = 0;//100000;
  //tauRoll = 0;//100000;
  //tauYaw = 0;//100000;
  //thrust = 0;//100000;
  
  thrust = 10000;
  thrust = thrust*0.6;

//  if(strcmp(name,"BLO") == 0) thrust = thrust*0.99;
  reset_PID_control_pitch();
  reset_PID_control_roll();
  reset_PID_control_yaw();  
  
  // Main loop
  while (wb_robot_step(timestep) != -1) {
    const double time = wb_robot_get_time();  // in seconds.

    // (Task-N) get the sensor values
    const double roll  =  wb_inertial_unit_get_roll_pitch_yaw(imu)[0] + M_PI / 2.0;
    const double pitch = -wb_inertial_unit_get_roll_pitch_yaw(imu)[1];
    const double yaw   =  wb_inertial_unit_get_roll_pitch_yaw(imu)[2] + M_PI/2.0;
    
    printf("roll = %f, pitch = %f, yaw = %f\n",roll,pitch,yaw);

    const double *gps_vals = wb_gps_get_values(gps);
    printf("gps[0] = %f, gps[1] = %f, gps[2] = %f\n",gps_vals[0],gps_vals[1],gps_vals[2]);

    const double z = wb_gps_get_values(gps)[1];
    const double roll_acceleration = wb_gyro_get_values(gyro)[0];
    const double pitch_acceleration = wb_gyro_get_values(gyro)[1];

    /*
    Kp = 1.0;
    Kd = 0.01;
    Ki = 0.07; 
    */
    
   dT = 5.0*timestep/1000.0;
       
    double pitchThrust =  200*do_PID_control_pitch(pitch, target_pitch, Kp, Ki, Kd, dT);
    double rollThrust =   100*do_PID_control_roll(roll, target_roll, Kp, Ki, Kd, dT);
    double yawThrust =    100*do_PID_control_yaw(yaw, target_yaw, Kp, Ki, Kd, dT);

/*
    bool heightControl = false;
    double Kp_1, Ki_1, Kd_1;
    
      if(gps_vals[1] > 20) {
        Kp_1 = 1.0;
        Kd_1 = 0.1;
        Ki_1 = 1.0;
        heightControl = true;
      }

    if(heightControl)
        thrust = 500*do_PID_control_1(gps_vals[1],25.0, Kp_1, Ki_1, Kd_1, dT);      
*/       
    printf("Time = %f, Thrusts:vert = %f, pitch = %f, roll = %f, yaw = %f\n",time,thrust,pitchThrust, rollThrust, yawThrust);
     
    thrustToRotorSpeed(thrust, pitchThrust, rollThrust, yawThrust, &FL, &RL, &RR, &FR);
     
    // Blink the front LEDs alternatively with a 1 second rate.
    const bool led_state = ((int)time) % 2;
    wb_led_set(front_left_led, led_state);
    wb_led_set(front_right_led, !led_state);

    // Stabilize the Camera by actuating the camera motors according to the gyro feedback.
    wb_motor_set_position(camera_roll_motor, -0.115 * roll_acceleration);
    wb_motor_set_position(camera_pitch_motor, -0.1 * pitch_acceleration);

    printf("FR = %f, FL = %f, RR = %f, RL = %f\n",FR,FL,RR,RL);

    // (Task-1) Indentify the rotors
    //FL = 0;
    //FR = 30;
    //RL = 0;
    //RR = 0;
    
    wb_motor_set_velocity(front_left_motor,  FL);
    wb_motor_set_velocity(front_right_motor,-FR);
    wb_motor_set_velocity(rear_left_motor,  -RL);
    wb_motor_set_velocity(rear_right_motor,  RR);
    
    int key = wb_keyboard_get_key();
    //printf(">>>>>>>>> key = %d\n",key);
     
    while (key > 0) {
      switch (key) {
        case WB_KEYBOARD_UP:
          target_pitch += 1.0*M_PI/180.0;
          printf("pitch = %f\n",target_pitch);
          break;
         case WB_KEYBOARD_DOWN:
          target_pitch -= 1.0*M_PI/180.0;
          printf("pitch = %f\n",target_pitch);          
          break;
         case (WB_KEYBOARD_SHIFT + WB_KEYBOARD_UP):
           thrust += 50;
           printf("thrust: %f\n", thrust);
           break;
         case (WB_KEYBOARD_SHIFT + WB_KEYBOARD_DOWN):
           thrust -= 50;
           printf("thrust: %f\n", thrust);                   
      }
      key = wb_keyboard_get_key();
    }
    
    //writeDataToOctave(logFile,time,pitch,roll,yaw,z);   

    printf("pitch = %f, target pitch = %f\n",pitch,target_pitch);
        
  }
  
  //writeOctaveTail(logFile);
  fflush(logFile);
  fclose(logFile);
  
  wb_robot_cleanup();

  return EXIT_SUCCESS;
}
